/* This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/** @brief library to communicate with the Maxim DS1307 I2C RTC IC (code)
 *  @file rtc_ds1307.c
 *  @author King Kévin <kingkevin@cuvoodoo.info>
 *  @date 2016
 *  @note peripherals used: I2C @ref rtc_ds1307_i2c, GPIO @ref rtc_ds1307_square_wave
 */

/* standard libraries */
#include <stdint.h> // standard integer types
#include <stdio.h> // standard I/O facilities
#include <stdlib.h> // general utilities

/* STM32 (including CM3) libraries */
#include <libopencm3/stm32/rcc.h> // real-time control clock library
#include <libopencm3/stm32/gpio.h> // general purpose input output library
#include <libopencm3/stm32/i2c.h> // I2C library
#include <libopencm3/cm3/nvic.h> // interrupt handler
#include <libopencmsis/core_cm3.h> // Cortex M3 utilities

#include "global.h" // global utilities
#include "rtc_ds1307.h" // RTC header and definitions

/** @defgroup rtc_ds1307_i2c I2C peripheral used to communication with the DS1307 RTC IC
 *  @{
 */
/** I2C peripheral */
#define I2C I2C1 /**< I2C peripheral */
#define I2C_RCC RCC_I2C1 /**< I2C peripheral clock */
#define I2C_PORT GPIOB /**< I2C peripheral port */
#define I2C_PIN_SDA GPIO_I2C1_SDA /**< I2C peripheral data pin (PB7) */
#define I2C_PIN_SCL GPIO_I2C1_SCL /**< I2C peripheral clock pin (PB6) */
/** @} */
#define I2C_ADDR 0x68 /**< DS1307 I2C address (fixed to 0b1101000) */

#if defined(SQUARE_WAVE_EXTI) && defined(SQUARE_WAVE_IRQ) && defined(SQUARE_WAVE_ISR) && defined(SQUARE_WAVE_HANDLING) && SQUARE_WAVE_HANDLING
volatile bool square_wave_flag = false;
#endif

void rtc_setup(void)
{
	/* enable peripheral */
	rcc_periph_clock_enable(RCC_AFIO); // enable clock for alternate function
	rcc_periph_clock_enable(I2C_RCC); // enable clock for I2C peripheral
	gpio_set_mode(I2C_PORT, GPIO_MODE_OUTPUT_10_MHZ, GPIO_CNF_OUTPUT_ALTFN_OPENDRAIN, I2C_PIN_SDA | I2C_PIN_SCL); // setup I2C pin

	/* configure I2C peripheral (see RM008 26.3.3, I2C master) */
	i2c_reset(I2C); // reset configuration
	i2c_peripheral_disable(I2C); // I2C needs to be disable to be configured
	i2c_set_clock_frequency(I2C, rcc_apb1_frequency/1E6); // configure the peripheral clock to the APB1 freq (where it is connected to)
	i2c_set_standard_mode(I2C); // the DS1307 has a maximum I2C SCL freq if 100 kHz (corresponding to the standard mode)
	i2c_set_ccr(I2C, rcc_apb1_frequency/(100E3*2)); // set Thigh/Tlow to generate frequency of 100 kHz
	i2c_set_trise(I2C, rcc_apb1_frequency/1E6); // max rise time for 100 kHz is 1000 ns (~1 MHz)
	i2c_peripheral_enable(I2C); // enable I2C after configuration completed

	/* setup square wave interrupt input */
#if defined(SQUARE_WAVE_RCC) && defined(SQUARE_WAVE_PORT) && defined(SQUARE_WAVE_PIN)
	rcc_periph_clock_enable(SQUARE_WAVE_RCC); // enable GPIO peripheral
	gpio_set_mode(SQUARE_WAVE_PORT, GPIO_MODE_INPUT, GPIO_CNF_INPUT_PULL_UPDOWN, SQUARE_WAVE_PIN); // set pin to input with pull up
	gpio_set(SQUARE_WAVE_PORT, SQUARE_WAVE_PIN); // pull up since the square wave output is open drain
#if defined(SQUARE_WAVE_EXTI) && defined(SQUARE_WAVE_IRQ) && defined(SQUARE_WAVE_ISR)
	rcc_periph_clock_enable(RCC_AFIO); // enable alternate function clock for external interrupt
	exti_select_source(SQUARE_WAVE_EXTI, SQUARE_WAVE_PORT); // mask external interrupt of this pin only for this port
	exti_set_trigger(SQUARE_WAVE_EXTI, EXTI_TRIGGER_FALLING); // trigger on falling
	exti_enable_request(SQUARE_WAVE_EXTI); // enable external interrupt
	nvic_enable_irq(SQUARE_WAVE_IRQ); // enable interrupt
#endif
#endif

}

/** @brief read memory from RTC IC
 *  @param[in] addr start address for memory to read
 *  @param[out] data buffer to store read memory
 *  @param[in] len number of byte to read from the memory
 *  @return if read succeeded
 */
static bool rtc_read_memory(uint8_t addr, uint8_t* data, size_t len)
{
	bool to_return = false; // return if read succeeded
	if (data==NULL || len==0) { // verify there it data to be read
		goto error;
	}
	i2c_send_start(I2C); // send start condition to start transaction
	while (!(I2C_SR1(I2C) & I2C_SR1_SB)); // wait until start condition is transmitted
	if (!(I2C_SR2(I2C) & I2C_SR2_MSL)) { // verify if in master mode
		goto error;
	}
	i2c_send_7bit_address(I2C, I2C_ADDR, I2C_WRITE); // select slave
	while (!(I2C_SR1(I2C) & I2C_SR1_ADDR)); // wait until address is transmitted
	if (!((I2C_SR2(I2C) & I2C_SR2_TRA))) { // verify we are in transmit mode (and read SR2 to clear ADDR)
		goto error;
	}
	i2c_send_data(I2C, addr); // send memory address we want to read
	while (!(I2C_SR1(I2C) & I2C_SR1_TxE)); // wait until byte has been transmitted
	i2c_send_start(I2C); // send restart condition to switch from write to read mode
	while (!(I2C_SR1(I2C) & I2C_SR1_SB)); // wait until start condition is transmitted
	i2c_send_7bit_address(I2C, I2C_ADDR, I2C_READ); // select slave
	while (!(I2C_SR1(I2C) & I2C_SR1_ADDR)); // wait until address is transmitted
	if ((I2C_SR2(I2C) & I2C_SR2_TRA)) { // verify we are in read mode (and read SR2 to clear ADDR)
		goto error;
	}
	for (size_t i=0; i<len; i++) { // read bytes
		if (i==len-1) { // prepare to sent NACK for last byte
			i2c_disable_ack(I2C); // NACK received to stop slave transmission
			i2c_send_stop(I2C); // send STOP after receiving byte
		} else {
			i2c_enable_ack(I2C); // ACK received byte to continue slave transmission
		}
		while (!(I2C_SR1(I2C) & I2C_SR1_RxNE)); // wait until byte has been received
		data[i] = i2c_get_data(I2C); // read received byte
	}
	to_return = true;
error:
	if (I2C_SR2(I2C) & I2C_SR2_BUSY) { // release bus if busy
		i2c_send_stop(I2C); // send stop to release bus
	}
	while (I2C_SR2(I2C) & I2C_SR2_MSL); // wait until bus released (non master mode)
	return to_return;
}

bool rtc_oscillator_disabled(void)
{
	uint8_t data[1] = {0}; // to read data over I2C
	rtc_read_memory(0, data, LENGTH(data)); // read a single byte containing CH value
	return data[0]&0x80; // return CH bit value to indicate if oscillator is disabled
}

uint16_t rtc_read_square_wave(void)
{
	uint16_t to_return = 0; // square wave frequency to return (in Hz)
	uint8_t data[1] = {0}; // to read data over I2C
	const uint16_t rtc_rs[] = {1, 4096, 8192, 32768}; // RS1/RS0 values
	rtc_read_memory(7, data, LENGTH(data)); // read a single byte containing control register
	if (data[0]&0x10) { // verify if the square wave is enabled (SQWE)
		to_return = rtc_rs[data[0]&0x03]; // read RS1/RS0 and get value
	} else {
		to_return = 0; // square wave output is disabled
	}
	return to_return;
}

uint8_t rtc_read_seconds(void)
{
	uint8_t to_return = 0; // seconds to return
	uint8_t data[1] = {0}; // to read data over I2C
	rtc_read_memory(0, data, LENGTH(data)); // read a single byte containing seconds value
	to_return = ((data[0]&0x70)>>4)*10+(data[0]&0x0f); // convert BCD coding into seconds
	return to_return;
}

uint8_t rtc_read_minutes(void)
{
	uint8_t to_return = 0; // minutes to return
	uint8_t data[1] = {0}; // to read data over I2C
	rtc_read_memory(1, data, LENGTH(data)); // read a single byte containing minutes value
	to_return = (data[0]>>4)*10+(data[0]&0x0f); // convert BCD coding into minutes
	return to_return;
}

uint8_t rtc_read_hours(void)
{
	uint8_t to_return = 0; // hours to return
	uint8_t data[1] = {0}; // to read data over I2C
	rtc_read_memory(2, data, LENGTH(data)); // read a single byte containing hours value
	if (data[0]&0x40) { // 12 hour mode
		if (data[0]&0x02) { // PM
			to_return += 12; // add the 12 hours
		}
		to_return += ((data[0]&0x10)>>4)*10; // convert BCD coding into hours (first digit)
	} else {
		to_return = ((data[0]&0x30)>>4)*10; // convert BCD coding into hours (first digit)
	}
	to_return += (data[0]&0x0f); // convert BCD coding into hours (second digit)
	return to_return;
}

uint8_t rtc_read_day(void)
{
	uint8_t to_return = 0; // day to return
	uint8_t data[1] = {0}; // to read data over I2C
	rtc_read_memory(3, data, LENGTH(data)); // read a single byte containing day value
	to_return = (data[0]&0x07); // convert BCD coding into days
	return to_return;
}

uint8_t rtc_read_date(void)
{
	uint8_t to_return = 0; // date to return
	uint8_t data[1] = {0}; // to read data over I2C
	rtc_read_memory(4, data, LENGTH(data)); // read a single byte containing date value
	to_return = ((data[0]&0x30)>>4)*10+(data[0]&0x0f); // convert BCD coding into date
	return to_return;
}

uint8_t rtc_read_month(void)
{
	uint8_t to_return = 0; // month to return
	uint8_t data[1] = {0}; // to read data over I2C
	rtc_read_memory(5, data, LENGTH(data)); // read a single byte containing month value
	to_return = ((data[0]&0x10)>>4)*10+(data[0]&0x0f); // convert BCD coding into month
	return to_return;
}

uint16_t rtc_read_year(void)
{
	uint16_t to_return = 2000; // year to return
	uint8_t data[1] = {0}; // to read data over I2C
	rtc_read_memory(6, data, LENGTH(data)); // read a single byte containing year value
	to_return += ((data[0]&0xf0)>>4)*10+(data[0]&0x0f); // convert BCD coding into year
	return to_return;
}

uint16_t* rtc_read_time(void)
{
	static uint16_t time[7] = {0}; // store time {seconds, minutes, hours, day, date, month, year}
	uint8_t data[7] = {0}; // to read data over I2C
	rtc_read_memory(0, data, LENGTH(data)); // read all time bytes
	time[0] = ((data[0]&0x70)>>4)*10+(data[0]&0x0f); // convert seconds from BCD
	time[1] = (data[1]>>4)*10+(data[1]&0x0f); // convert minutes from BCD
	time[2] = 0; // re-initialize hours
	if (data[2]&0x40) { // 12 hour mode
		if (data[2]&0x02) { // PM
			time[2] += 12; // add the 12 hours
		}
		time[2] += ((data[2]&0x10)>>4)*10; // convert BCD coding into hours (first digit)
	} else {
		time[2] = ((data[2]&0x30)>>4)*10; // convert BCD coding into hours (first digit)
	}
	time[2] += (data[2]&0x0f); // convert BCD coding into hours (second digit)
	time[3] = (data[3]&0x07); // convert BCD coding into days
	time[4] = ((data[4]&0x30)>>4)*10+(data[4]&0x0f); // convert BCD coding into date
	time[5] = ((data[5]&0x10)>>4)*10+(data[5]&0x0f); // convert BCD coding into month
	time[6] = 2000+((data[6]&0xf0)>>4)*10+(data[6]&0x0f); // convert BCD coding into year
	return time;
}

/** @brief write memory into RTC IC
 *  @param[in] addr start address for memory to be written
 *  @param[in] data buffer to for memory to be written
 *  @param[in] len number of byte to write into the memory
 *  @return if write succeeded
 */
static bool rtc_write_memory(uint8_t addr, uint8_t* data, size_t len)
{
	bool to_return = false; // return if read succeeded
	if (data==NULL || len==0) { // verify there it data to be read
		goto error;
	}
	i2c_send_start(I2C); // send start condition to start transaction
	while (!(I2C_SR1(I2C) & I2C_SR1_SB)); // wait until start condition is transmitted
	if (!(I2C_SR2(I2C) & I2C_SR2_MSL)) { // verify if in master mode
		goto error;
	}
	i2c_send_7bit_address(I2C, I2C_ADDR, I2C_WRITE); // select slave
	while (!(I2C_SR1(I2C) & I2C_SR1_ADDR)); // wait until address is transmitted
	if (!((I2C_SR2(I2C) & I2C_SR2_TRA))) { // verify we are in transmit mode (and read SR2 to clear ADDR)
		goto error;
	}
	i2c_send_data(I2C, addr); // send memory address we want to read
	while (!(I2C_SR1(I2C) & I2C_SR1_TxE)); // wait until byte has been transmitted
	for (size_t i=0; i<len; i++) { // write bytes
		i2c_send_data(I2C, data[i]); // send byte to be written in memory
		while (!(I2C_SR1(I2C) & I2C_SR1_TxE)); // wait until byte has been transmitted
	}
	to_return = true;
error:
	if (I2C_SR2(I2C) & I2C_SR2_BUSY) { // release bus if busy
		i2c_send_stop(I2C); // send stop to release bus
	}
	while (I2C_SR2(I2C) & I2C_SR2_MSL); // wait until bus released (non master mode)
	return to_return;
}

bool rtc_oscillator_disable(void)
{
	uint8_t data[1] = {0}; // to write CH value data over I2C
	rtc_read_memory(0, data, LENGTH(data)); // read seconds with CH value
	data[0] |= 0x80; // set CH to disable oscillator
	return rtc_write_memory(0, data, LENGTH(data)); // write current seconds with CH value
}

bool rtc_oscillator_enable(void)
{
	uint8_t data[1] = {0}; // to write CH value data over I2C
	rtc_read_memory(0, data, LENGTH(data)); // read seconds with CH value
	data[0] &= 0x7f; // clear CH to enable oscillator
	return rtc_write_memory(0, data, LENGTH(data)); // write current seconds with CH value
}

bool rtc_write_square_wave(uint16_t frequency)
{
	uint8_t data[1] = {0}; // to write control register value data over I2C
	switch (frequency) { // set RS1/RS0 based on frequency
		case 0:
			data[0] = 0;
			break;
		case 1:
			data[0] = 0|(1<<4);
			break;
		case 4096:
			data[0] = 1|(1<<4);
			break;
		case 8192:
			data[0] = 2|(1<<4);
			break;
		case 32768:
			data[0] = 3|(1<<4);
			break;
		default: // unspecified frequency
			return false;
	}
	return rtc_write_memory(7, data, LENGTH(data)); // write current seconds with CH value
}

bool rtc_write_seconds(uint8_t seconds)
{
	if (seconds>59) {
		return false;
	}
	uint8_t data[1] = {0}; // to read CH value data and write seconds value over I2C
	if (!rtc_read_memory(0, data, LENGTH(data))) { // read seconds with CH value
		return false;
	}
	data[0] &= 0x80; // only keep CH flag
	data[0] |= (((seconds/10)%6)<<4)+(seconds%10); // encode seconds in BCD format
	return rtc_write_memory(0, data, LENGTH(data)); // write current seconds with previous CH value
}

bool rtc_write_minutes(uint8_t minutes)
{
	if (minutes>59) {
		return false;
	}
	uint8_t data[1] = {0}; // to write time value
	data[0] = (((minutes/10)%6)<<4)+(minutes%10); // encode minutes in BCD format
	return rtc_write_memory(1, data, LENGTH(data)); // write time value on RTC
}

bool rtc_write_hours(uint8_t hours)
{
	if (hours>24) {
		return false;
	}
	uint8_t data[1] = {0}; // to write time value
	data[0] = (((hours/10)%3)<<4)+(hours%10); // encode hours in BCD 24h format
	return rtc_write_memory(2, data, LENGTH(data)); // write time value on RTC
}

bool rtc_write_day(uint8_t day)
{
	if (day<1 || day>7) {
		return false;
	}
	uint8_t data[1] = {0}; // to write time value
	data[0] = (day%8); // encode day in BCD format
	return rtc_write_memory(3, data, LENGTH(data)); // write time value on RTC
}

bool rtc_write_date(uint8_t date)
{
	if (date<1 || date>31) {
		return false;
	}
	uint8_t data[1] = {0}; // to write time value
	data[0] = (((date/10)%4)<<4)+(date%10); // encode date in BCD format
	return rtc_write_memory(4, data, LENGTH(data)); // write time value on RTC
}

bool rtc_write_month(uint8_t month)
{
	if (month<1 || month>12) {
		return false;
	}
	uint8_t data[1] = {0}; // to write time value
	data[0] = (((month/10)%2)<<4)+(month%10); // encode month in BCD format
	return rtc_write_memory(5, data, LENGTH(data)); // write time value on RTC
}

bool rtc_write_year(uint16_t year)
{
	if (year<2000 || year>2099) {
		return false;
	}
	uint8_t data[1] = {0}; // to write time value
	data[0] = (((year/10)%10)<<4)+(year%10); // encode year in BCD format
	return rtc_write_memory(6, data, LENGTH(data)); // write time value on RTC
}

bool rtc_write_time(uint8_t seconds, uint8_t minutes, uint8_t hours, uint8_t day, uint8_t date, uint8_t month, uint16_t year)
{
	uint8_t data[7] = {0}; // to write all time values
	// seconds
	if (seconds>59) {
		return false;
	}
	if (!rtc_read_memory(0, data, 1)) { // read seconds with CH value
		return false;
	}
	data[0] &= 0x80; // only keep CH flag
	data[0] |= (((seconds/10)%6)<<4)+(seconds%10); // encode seconds in BCD format
	// minutes
	if (minutes>59) {
		return false;
	}
	data[1] = (((minutes/10)%6)<<4)+(minutes%10); // encode minutes in BCD format
	// hours
	if (hours>24) {
		return false;
	}
	data[2] = (((hours/10)%3)<<4)+(hours%10); // encode hours in BCD 24h format
	// day
	if (day<1 || day>7) {
		return false;
	}
	data[3] = (day%8); // encode day in BCD format
	// date
	if (date<1 || date>31) {
		return false;
	}
	data[4] = (((date/10)%4)<<4)+(date%10); // encode date in BCD format
	// month
	if (month<1 || month>12) {
		return false;
	}
	data[5] = (((month/10)%2)<<4)+(month%10); // encode month in BCD format
	// year
	if (year<2000 || year>2099) {
		return false;
	}
	data[6] = (((year/10)%10)<<4)+(year%10); // encode year in BCD format

	return rtc_write_memory(0, data, LENGTH(data)); // write time values on RTC
}

#if defined(SQUARE_WAVE_EXTI) && defined(SQUARE_WAVE_IRQ) && defined(SQUARE_WAVE_ISR) && defined(SQUARE_WAVE_HANDLING) && SQUARE_WAVE_HANDLING
/** @brief square wave input interrupt */
void SQUARE_WAVE_ISR(void)
{
	exti_reset_request(SQUARE_WAVE_EXTI); // reset interrupt
	square_wave_flag = true; // update flag
}
#endif

